/*
** BeNES - Nintendo Entertaiment System Emulator for BeOS
**
** * 1999 by makoto yamagata
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
//
// custom 6502 CPU core
//
#include <stdlib.h>
#include <string.h>
#include "NesCPU.h"
#include "NesIO.h"

const int CPU_SLICE_CYCLES = 113;
const int CPU_UNIT_CYCLES = 7457;
const int CPU_FRAME_CYCLES = CPU_UNIT_CYCLES * 4;

//============================================ CPU Control ===
/*---------------------------------------------- reset ---*/
void NesCPU::reset()
{
	clearRam();
	memset(&m6502, 0, sizeof(m6502));
	m6502.IPeriod = CPU_SLICE_CYCLES;
	m6502.User = reinterpret_cast<void*>(this);
	m6502.Loop = Loop;
	m6502.Ram = mRam;
	setReadFunction();
	setWriteFunction();
	mIO->connect(this);
	mIO->ROM_Init(mRam);
	mFrameIrqEnabled = false;
	mFrameClock = 0;
	mTicks = 0;
	Reset6502(&m6502);
	mIO->APU_Reset();
}

/*--------------------------------------------- execute ---*/
void NesCPU::run()
{
	mIsRunning = true;
	Run6502(&m6502);
	mIO->ROM_UnInit(mRam);
}

/*------------------------------------- RAM initialize ---*/
void NesCPU::clearRam()
{
	memset(mRam, 0, sizeof(mRam));
}

/*---------------------------------------- Set Read Func ---*/
void NesCPU::setReadFunction()
{
	// Read RAM
	m6502.Read[0] = 
	m6502.Read[1] = ReadRam;
	// Read PPU Register
	m6502.Read[2] = 
	m6502.Read[3] = ReadPPURegisters;
	// APU & SPRITE DMA & JOYSTICK
	m6502.Read[4] = Read0x4000;
	// Expansion Modules  
	m6502.Read[5] = ReadNormal;
	// SRAM
	m6502.Read[6] = 
	m6502.Read[7] = ReadNormal;
	// ROM
	m6502.Read[8] = 
	m6502.Read[9] =
	m6502.Read[10] = 
	m6502.Read[11] =
	m6502.Read[12] = 
	m6502.Read[13] =
	m6502.Read[14] = 
	m6502.Read[15] = ReadNormal;
}

/*--------------------------------------- Set Write Func ---*/
void NesCPU::setWriteFunction()
{
	// Write RAM
	m6502.Write[0] = 
	m6502.Write[1] = WriteRam;
	// Write PPU Register
	m6502.Write[2] = 
	m6502.Write[3] = WritePPURegisters;
	// APU & SPRITE DMA & JOYSTICK
	m6502.Write[4] = Write0x4000;
	// Expansion Modules  
	m6502.Write[5] = WriteNormal;
	// SRAM
	m6502.Write[6] = 
	m6502.Write[7] = WriteNormal;
	// ROM
	m6502.Write[8] = 
	m6502.Write[9] =
	m6502.Write[10] = 
	m6502.Write[11] =
	m6502.Write[12] = 
	m6502.Write[13] =
	m6502.Write[14] = 
	m6502.Write[15] = WriteROM;
}

//============================================================
/*--------------------------------------------- Loop6502 ---*/
uchar NesCPU::Loop(void* user)
{
	NesCPU* cpu = reinterpret_cast<NesCPU*>(user);
	uchar reply = INT_NONE;
	if (cpu->mIsRunning) {
		cpu->mIO->PPU_ScanLine();
		cpu->mFrameClock += (CPU_SLICE_CYCLES-cpu->m6502.ICount);
		if ((cpu->mFrameClock>CPU_FRAME_CYCLES) && cpu->mFrameIrqEnabled) {
			cpu->mFrameClock %= CPU_FRAME_CYCLES;
			cpu->intIRQ();
			cpu->mRam[0x4015] |= 0x40;
		}
		cpu->mTicks += (CPU_SLICE_CYCLES-cpu->m6502.ICount);
		if (cpu->mTicks > CPU_UNIT_CYCLES) {
			cpu->mTicks %= CPU_UNIT_CYCLES;
			cpu->mIO->APU_StatusChange();
		}
	}
	else {
		reply = INT_QUIT;
	}
	return reply;
}

//================================================= RAM IO ===
/*----------------------------------------- Read CPU RAM ---*/
uchar NesCPU::ReadRam(void* user, int address)
{
	NesCPU* cpu = reinterpret_cast<NesCPU*>(user);
	return cpu->mRam[address];
}

/*---------------------------------------- Write CPU RAM ---*/
void NesCPU::WriteRam(void* user, int address, uchar value)
{
	NesCPU* cpu = reinterpret_cast<NesCPU*>(user);
	cpu->mRam[address^0x0000] = value;
	cpu->mRam[address^0x0800] = value;
	cpu->mRam[address^0x1000] = value;
	cpu->mRam[address^0x1800] = value;
}

//======================================== PPU REGISTER IO ===
/*------------------------------------ Read PPU REGISTER ---*/
uchar NesCPU::ReadPPURegisters(void* user, int address)
{
	NesCPU* cpu = reinterpret_cast<NesCPU*>(user);
	uchar ret = 0xff;
	switch (address&0x2007) {
	case 0x2002:
		ret = cpu->mIO->PPU_ReadStatus();
		break;
	case 0x2004:
		ret = cpu->mIO->PPU_ReadSpriteRam();
		break;
	case 0x2007:
		ret = cpu->mIO->PPU_ReadPPURam();
		break;
	default:
		ret = cpu->mRam[address];
		break;
	}
	return ret;
}

/*----------------------------------- Write PPU REGISTER ---*/
void NesCPU::WritePPURegisters(
	void* user, int address, uchar value)
{
	NesCPU* cpu = reinterpret_cast<NesCPU*>(user);
	switch (address&0x2007) {
	case 0x2000:
		cpu->mIO->PPU_WriteControlA(value);
		break;
	case 0x2001:
		cpu->mIO->PPU_WriteControlB(value);
		break;
	case 0x2003:
		cpu->mIO->PPU_WriteSpriteRamAddress(value);
		break;
	case 0x2004:
		cpu->mIO->PPU_WriteSpriteRam(value);
		break;
	case 0x2005:
		cpu->mIO->PPU_WriteBackgroundScroll(value);
		break;
	case 0x2006:
		cpu->mIO->PPU_WritePPURamAddress(value);
		break;
	case 0x2007:
		cpu->mIO->PPU_WritePPURam(value);
		break;
	default:
		cpu->mRam[address] = value;
		break;
	}
}

//================================= APU,JOY and SPRITE DMA ===
/*----------------------------------------------------------*/
uchar NesCPU::Read0x4000(void* user, int address)
{
	NesCPU* cpu = reinterpret_cast<NesCPU*>(user);
	uchar ret;
	if (address == 0x4015) {
		cpu->mRam[0x4015] &= ~0x0f;
		cpu->mRam[0x4015] |= cpu->mIO->APU_Read4015();
		ret = cpu->mRam[0x4015];
		cpu->mRam[0x4015] &= ~0x40;
		cpu->mFrameClock = 0;
	}else if (address == 0x4016) { // Joypad A
		cpu->mRam[0x4016] = cpu->mIO->JOY_Read();
		ret = cpu->mRam[0x4016];
	}else if (address == 0x4017) { // Joypad B Not Support...
		ret = 0x40;
	}else { // APU Not Support...
		ret = cpu->mRam[address];
	}
	return ret;
}

/*----------------------------------------------------------*/
inline void NesCPU::WriteAPURegisters(
	NesCPU* cpu, int address, uchar value)
{
	switch (address) {
	case 0x4000:
	case 0x4001:
	case 0x4002:
	case 0x4003:
		cpu->mIO->APU_WriteSq0(address-0x4000, value);
		break;
	case 0x4004:
	case 0x4005:
	case 0x4006:
	case 0x4007:
		cpu->mIO->APU_WriteSq1(address-0x4004, value);
		break;
	case 0x4008:
	case 0x4009:
	case 0x400A:
	case 0x400B:
		cpu->mIO->APU_WriteTri(address-0x4008, value);
		break;
	case 0x400C:
	case 0x400D:
	case 0x400E:
	case 0x400F:
		cpu->mIO->APU_WriteNiz(address-0x400C, value);
		break;
	}
}

/*----------------------------------------------------------*/
void NesCPU::Write0x4000(void* user, int address, uchar value)
{
	NesCPU* cpu = reinterpret_cast<NesCPU*>(user);
	if (address <= 0x400F) {
		WriteAPURegisters(cpu, address, value);
	}else if (address == 0x4014) { // sprite DMA
		cpu->mIO->PPU_SpriteDMA(cpu->mRam, value);
	}else if (address == 0x4015) {
		cpu->mRam[0x4015] = value;
		cpu->mIO->APU_Write4015(value);
		if (value & 0x10) {
			cpu->mRam[0x4015] &= ~0x80;
		}
	}else if (address == 0x4016) { // Joypad A
		if ((cpu->mRam[0x4016]&0x01) && (value==0)) {
			cpu->mIO->JOY_Reset();
		}
		cpu->mRam[address] = value;
	}else if (address == 0x4017) { // Joypad B Not Support...
		cpu->mRam[0x4017] = value;
		cpu->mFrameClock = 0;
		if (!(value & 0xc0)) {
			cpu->mFrameIrqEnabled = true;
		}
		else {
			cpu->mFrameIrqEnabled = false;
		}
	}else { // APU Not Support...
		cpu->mRam[address] = value;
	}
}

//========================================== ROM BANK SET ===
/*----------------------------------------------------------*/
void NesCPU::WriteROM(void* user, int address, uchar value)
{
	NesCPU* cpu = reinterpret_cast<NesCPU*>(user);
	cpu->mIO->ROM_Write(address, value);
}

//===================================== NORMAL RAM ACCESS ===
/*----------------------------------------------------------*/
uchar NesCPU::ReadNormal(void* user, int address)
{
	NesCPU* cpu = reinterpret_cast<NesCPU*>(user);
	return cpu->mRam[address];
}

/*----------------------------------------------------------*/
void NesCPU::WriteNormal(void* user, int address, uchar value)
{
	NesCPU* cpu = reinterpret_cast<NesCPU*>(user);
	cpu->mRam[address] = value;
}
