/*
** BeNES - Nintendo Entertaiment System Emulator for BeOS
**
** * 1999 by makoto yamagata
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
//
// Picture Processor Unit (PPU)
//
#include <string>
#include "NesPPU.h"
#include "NesRenderInfo.h"

struct SpriteCell {
	uchar y;
	uchar tile_index;
	uchar attr;
	uchar x;
public:
	bool isBackgroundFront() { return ((attr&0x20) != 0); }
	bool isHorizontalFlip() { return ((attr&0x40) != 0); }
	bool isVerticalFlip() { return ((attr&0x80) != 0); }
	uchar colorHi() { return (attr&0x03)<<2; }
};

static NesScreenRenderInfo ScreenInfo;

const int EXECUTE_NMI_LINE = 245;
const int VBLANK_LINE = 243;
const int DRAWBEGIN_LINE = 8;

//----------------------------------------------------------------
void NesPPU::init()
{
	mFrameCounter = 0;
	NesPPUImpl::init();
}

//-----------------------------------------------------------------
void NesPPU::renderLine(int scanline)
{
	int tiletbl_base = mControlA.screenPatternTableAddress();
	int nametbl_base = mControlA.nameTableAddress();
	uchar v_scroll = mBackgroundScroll.getVScrollValue();
	uchar h_scroll = mBackgroundScroll.getHScrollValue();
	int top = scanline + v_scroll;
	if (top >= 240) {
		top -= 240;
		nametbl_base ^= 0x0800;
	}
	int y = scanline;
	int x = 0;
	int width = 8 - (h_scroll & 0x07);
	bool firstBankChange = true;
	while (x < 256) {
		int left = x + h_scroll;
		if (left >= 256) {
			left -= 256;
			if (firstBankChange) {
				nametbl_base ^= 0x0400;
				firstBankChange = false;
			}
		}
		int nametbl = ScreenInfo.nametbl(nametbl_base, top, left);
		int value = static_cast<int>(mPPURam[nametbl]);
		int tiletbl = tiletbl_base + (value*16) + (top&0x07);
		uchar tileA = mPPURam[tiletbl];
		uchar tileB = mPPURam[tiletbl+8];
		int attr_tbl = ScreenInfo.attrtbl(nametbl_base+0x3c0, top, left);
		uchar attr = mPPURam[attr_tbl];
		uchar color_hi = (((attr>>ScreenInfo.shift(top, left))&03)<<2);
		for (int j=0; j<width; j++) {
			uchar color_index = 0;
			if ((tileA!=0) || (tileB!=0)) {
				color_index = color_hi|(tileB&0x01)*2 + (tileA&0x01);
			}
			tileA >>= 1;
			tileB >>= 1;
			int pos_x = x + ((width-1)-j);
			if (pos_x >= 256) continue;
			mScreenMask.bits[y][pos_x] = color_index;
			color_index = mPPURam[0x3F00+color_index];
			mVirtualScreen.bits[y][pos_x] = color_index;
		}
		x += width;
		width = 8;
	}
}

//-----------------------------------------------------------------
void NesPPU::putSprite(int scanline)
{
	SpriteCell* cell = reinterpret_cast<SpriteCell*>(mSpriteRam);

	ushort tiletbl_base = mControlA.spritePatternTableAddress();
	int spriteHeight = mControlA.is8x8Sprite()? 8: 16;
	
	for (int i=0; i<64; i++) {
		if ((cell[i].y+spriteHeight) != scanline) continue;
		ushort value = static_cast<ushort>(cell[i].tile_index);
		if (spriteHeight == 16) {
			tiletbl_base = 0x0000;
			if (value & 0x01) {
				value &= 0xfe;
				value |= 0x100;
			}
		}
		uchar color_hi = cell[i].colorHi();
		for (int line=0; line<spriteHeight; line++) {
			ushort tiletbl = tiletbl_base + (value*16) + (line&0x07) + ((line/8)*16);
			uchar tileA = mPPURam[tiletbl];
			uchar tileB = mPPURam[tiletbl+8];
			if ((tileA==0x00) && (tileB==0x00)) continue;
			for (int j=0; j<8; j++) {
				uchar color_index = (tileB&0x01)*2+(tileA&0x01);
				tileA >>= 1;
				tileB >>= 1;
				if (color_index == 0) continue;

				int x = cell[i].x;
				if (cell[i].isHorizontalFlip()) x += (j);
				else x += (7-j);
				if ((x>=256) || (x<0)) continue;
				
				int y = cell[i].y+1;
				if (cell[i].isVerticalFlip()) y += ((spriteHeight-1)-line);
				else y += line;
				if ((y>=240) || (y<0)) continue;

				color_index = color_hi|color_index;
				if (cell[i].isBackgroundFront() && (mScreenMask.bits[y][x]&~0x80)) {
					if (color_index) mScreenMask.bits[y][x] |= 0x80;
					mVirtualScreen.bits[y][x] = 
						mPPURam[0x3f00+(mScreenMask.bits[y][x]&~0x80)];
					continue;
				}
				if (!(cell[i].isBackgroundFront()) || (mScreenMask.bits[y][x]==0)) {
					if (!(mScreenMask.bits[y][x]&0x80)) {
					if (color_index) mScreenMask.bits[y][x] |= 0x80;
						color_index = mPPURam[0x3f10+color_index];
						mVirtualScreen.bits[y][x] = color_index;
					}
				}
			}
		}
	}
}
//-----------------------------------------------------------------
int NesPPU::spriteHit0y()
{
	SpriteCell* cell = reinterpret_cast<SpriteCell*>(mSpriteRam);
	ushort tiletbl_base = mControlA.spritePatternTableAddress();
	int spriteHeight = mControlA.is8x8Sprite()? 8: 16;
	ushort value = static_cast<ushort>(cell->tile_index);
	if (spriteHeight == 16) {
		tiletbl_base = 0x0000;
		if (value & 0x01) {
			value &= 0xfe;
			value |= 0x100;
		}
	}
	int add_y = 0;
	for (int line=0; line<spriteHeight; line++) {
		ushort tiletbl = tiletbl_base + (value*16) + (line&0x07) + ((line/8)*16);
		uchar tileA = mPPURam[tiletbl];
		uchar tileB = mPPURam[tiletbl+8];
		if ((tileA==0x00) && (tileB==0x00)) {
			add_y++;
		}
		else {
			break;
		}
		
	}
	return static_cast<int>(cell->y)+1+add_y;
}

//-----------------------------------------------------------------
int NesPPU::scanLine(NesCPUInterface* cpu)
{
	if (getScanline() == 0) {
		mStatus.setVBlankBit(false);
		mSpriteHitLine = spriteHit0y();
	}
	if (getScanline() == DRAWBEGIN_LINE) {
		NesPPUImpl::syncValue();
	}
	if ((getScanline()==mSpriteHitLine) && mControlB.isSpriteVisible()) {
		mStatus.setSpriteHitBit(true);
	}
	if ((getScanline()>=DRAWBEGIN_LINE) && (getScanline()<240)) {
		if (mFrameCounter == 0) {
			if (mControlB.isBgVisible()) {
				renderLine(getScanline());
			}
			else {
				memset(mVirtualScreen.bits[getScanline()], 
					0x0f, sizeof(mVirtualScreen.bits[0]));
				memset(mScreenMask.bits[getScanline()], 
					0x00, sizeof(mScreenMask.bits[0]));
			}
			if (mControlB.isSpriteVisible()) {
				putSprite(getScanline());
			}
		}
	}
	if (getScanline() == VBLANK_LINE) {
		mDisplay->updateScreen(&mVirtualScreen, (mFrameCounter==0));
		mFrameCounter = (mFrameCounter)? --mFrameCounter: mSkipFrames;
		mStatus.setVBlankBit(true);
		mStatus.setSpriteHitBit(false);
		mBackgroundScroll.resetToggle();
	}
	if (getScanline() == EXECUTE_NMI_LINE) {
		if (mControlA.executeNMIonVBlank()) {
			cpu->intNMI();
		}
	}
	nextScan();
	return getScanline();
}
