/*
** BeNES - Nintendo Entertaiment System Emulator for BeOS
**
** * 1999 by makoto yamagata
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifndef NESPPUIMPL_H_DEFINED
#define NESPPUIMPL_H_DEFINED

#include "NesTypes.h"
#include "NesPPURegisters.h"
#include "NesPPUInterface.h"
#include "NesPPURam.h"

const int PPU_SCANLINE_PAR_FRAME = 262;

class NesPPUImpl: public NesPPUInterface {
	ushort mScanline;
	uchar mPPURamLatch;
protected:
	NesPPUControlA mControlA;
	NesPPUControlB mControlB;
	NesPPUStatus mStatus;
	NesPPUSpriteRamAddress mSpriteRamAddress;
	NesPPUBackgroundScroll mBackgroundScroll;
	NesPPURamAddress mPPURamAddress;
	NesPPURam mPPURam;
	uchar mSpriteRam[0x100];
public:
	NesPPUImpl(): NesPPUInterface(), 
		mScanline(0),
		mPPURamLatch(0) { ; }
	void nextScan() {
		mScanline = (mScanline+1) % PPU_SCANLINE_PAR_FRAME;
	}
	ushort getScanline() { return mScanline; }
	void syncValue() {
		mControlA.syncRegister();
		mBackgroundScroll.syncRegister();
	}
	void resetScanline() { mScanline = 0; }
	void init() {
		resetScanline();
		mPPURam.init();
		memset(mSpriteRam, 0, sizeof(mSpriteRam));
	}
public:
	virtual bool isDrawing() {
		return ((mControlB.isBgVisible()) && (mScanline<240));
	}
	virtual void setMirrorType(int mirror) {
		switch (mirror) {
		case PPU_MIRROR_VERTICAL: 
			mPPURam.mirrorVertical();
			break;
		case PPU_MIRROR_HORIZONTAL:
			mPPURam.mirrorHorizontal();
			break;
		case PPU_MIRROR_ONELOW:
			mPPURam.mirrorOneLow();
			break;
		case PPU_MIRROR_ONEHIGH:
			mPPURam.mirrorOneHigh();
			break;
		}
	}
	virtual void memmove(size_t offset, uchar* mem, size_t size) {
		memcpy(&(mPPURam[offset]), mem, size);
	}
public:
	// 0x2000
	void writeControlA(uchar val) { mControlA = val; }
	// 0x2001
	void writeControlB(uchar val) { mControlB = val; }
	// 0x2002
	uchar readStatus() {
		uchar ret = mStatus();
		mStatus.setVBlankBit(false);
		mBackgroundScroll.resetToggle(); 
		mPPURamAddress.resetToggle();
		return ret;
	}
	// 0x2003
	void writeSpriteRamAddress(uchar val) { mSpriteRamAddress = val; }
	// 0x2004
	void writeSpriteRam(uchar value) {
		ushort address = mSpriteRamAddress();
		mSpriteRam[address] = value;
		mSpriteRamAddress += 1;
	}
	uchar readSpriteRam() {
		uchar ret = mSpriteRam[mSpriteRamAddress()];
		mSpriteRamAddress += 1;
		return ret;
	}
	// 0x2005
	void writeBackgroundScroll(uchar val) { mBackgroundScroll = val ; }
	// 0x2006
	void writePPURamAddress(uchar val) { 
		mPPURamAddress = val;
		ushort address = mPPURamAddress();
		mControlA.mirrorBit(address);
		mBackgroundScroll.mirrorBit(address, mScanline);
	}
	// 0x2007
	uchar readPPURam() {
		uchar ret = mPPURamLatch;
		mPPURamLatch = mPPURam[mPPURamAddress()];
		ushort inc = mControlA.ppuAddressIncrementValue();
		mPPURamAddress += inc;
		return ret;
	}
	void writePPURam(uchar value) {
		ushort address = mPPURamAddress();
		mPPURam[address] = value;
		if ((address&0x3fef) == 0x3f00) {
			// Image palette(0x3f00) and 
			// Sprite palette(0x3f10) mirroring.
			mPPURam[0x3f00] = mPPURam[0x3f04] = 
			mPPURam[0x3f08] = mPPURam[0x3f0c] =
			mPPURam[0x3f10] = mPPURam[0x3f14] = 
			mPPURam[0x3f18] = mPPURam[0x3f1c] = value;
		}
		ushort inc = mControlA.ppuAddressIncrementValue();
		mPPURamAddress += inc;
	}
	// 0x4014
	void spriteDMA(uchar* memory) {
		memcpy(mSpriteRam, memory, 0x100);
	}
};

#endif
